import '../../css/widget.css';
import { DOMWidgetView } from '@jupyter-widgets/base';
import { VBoxView } from '@jupyter-widgets/controls';
import * as d3 from 'd3';
import * as nurbs from 'nurbs';
import { debounce } from './utilities';
export class ParameterEditorView extends VBoxView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-parameter-editor');
    }
}
export class _EditorView extends VBoxView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-editor');
    }
}
export class BoolEditorView extends _EditorView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-bool-editor');
    }
}
export class IntEditorView extends _EditorView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-int-editor');
    }
}
export class FloatEditorView extends _EditorView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-float-editor');
    }
}
export class StringEditorView extends _EditorView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-string-editor');
    }
}
export class MaterialEditorView extends _EditorView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-material-editor');
    }
}
export class CurveEditorView extends _EditorView {
    initialize(parameters) {
        super.initialize(parameters);
        this.pWidget.addClass('pgl-curve-editor');
    }
}
export class _CurveEditorView extends DOMWidgetView {
    constructor() {
        super(...arguments);
        this.width = 250;
        this.height = 250;
        this.margin = 15;
        this.name = '';
        this.isFunction = false;
        this.controlPoints = [];
    }
    initialize(parameters) {
        super.initialize(parameters);
    }
    render() {
        super.render();
        this.pWidget.addClass('pgl-_curve-editor');
        const svg = d3.select(this.el)
            .append('svg')
            .classed('pgl-curve-editor-svg', true)
            .style('width', this.width)
            .style('height', this.height);
        this.name = this.model.get('name');
        this.isFunction = this.model.get('is_function');
        this.controlPoints = this.model.get('control_points').slice();
        this.curveType = this.model.get('curve_type');
        const [minx, miny, maxx, maxy] = this.controlPoints.reduce((m, p) => {
            m[0] = p[0] < m[0] ? p[0] : m[0];
            m[1] = p[1] < m[1] ? p[1] : m[1];
            m[2] = p[0] > m[2] ? p[0] : m[2];
            m[3] = p[1] > m[3] ? p[1] : m[3];
            return m;
        }, [Infinity, Infinity, -Infinity, -Infinity]);
        const domain = [Math.min(minx, miny), Math.max(maxx, maxy)];
        const xScale = d3.scaleLinear()
            .domain(domain)
            .range([this.margin, this.width - this.margin]);
        const yScale = d3.scaleLinear()
            .domain(domain)
            .range([this.height - this.margin, this.margin]);
        const lineGenerator = d3.line()
            .x(d => xScale(d[0]))
            .y(d => yScale(d[1]))
            .curve(d3.curveLinear);
        const dx = ((this.width / 2) - (xScale(maxx) + xScale(minx)) / 2);
        const dy = ((this.height / 2) - (yScale(maxy) + yScale(miny)) / 2);
        const g = svg.append('g')
            .attr('transform', `translate(${dx},${dy})`);
        g.append('path')
            .classed('grid', true)
            .attr('d', lineGenerator([[0, Math.min(minx, miny)], [0, Math.max(maxx, maxy)]]));
        g.append('path')
            .classed('grid', true)
            .attr('d', lineGenerator([[Math.min(minx, miny), 0], [Math.max(maxx, maxy), 0]]));
        const ctrlPath = g.append('path').classed('line', true);
        const curvePath = g.append('path').classed('curve', true);
        const circles = g.selectAll('.point')
            .data(this.controlPoints)
            .enter()
            .append('circle')
            .classed('point', true)
            .attr('r', 10)
            .attr('cx', d => xScale(d[0]))
            .attr('cy', d => yScale(d[1]));
        this.onControlPointsChanged(g, dx, dy, xScale, yScale, ctrlPath, curvePath, circles, lineGenerator);
        this.listenTo(this.model, 'change:control_points', () => {
            const controlPoints = this.model.get('control_points');
            if (controlPoints !== this.controlPoints) {
                this.controlPoints = controlPoints.slice();
                this.onControlPointsChanged(g, dx, dy, xScale, yScale, ctrlPath, curvePath, circles, lineGenerator);
            }
        });
    }
    onControlPointsChanged(g, dx, dy, xScale, yScale, ctrlPath, curvePath, circles, lineGenerator) {
        const isFunction = this.isFunction;
        const width = this.width;
        const height = this.height;
        const margin = this.margin;
        const draw = () => {
            switch (this.curveType) {
                case "NurbsCurve2D" /* NURBS */:
                    ctrlPath.attr('d', lineGenerator(this.controlPoints));
                    const curve = nurbs({
                        points: this.controlPoints,
                        degree: 3,
                        boundary: 'clamped'
                    });
                    const x0 = curve.domain[0][0];
                    const x1 = curve.domain[0][1];
                    const n = (width + height) / 4;
                    const curveSamples = [];
                    for (let i = 0; i < n; i++) {
                        const t0 = x0 + ((x1 - x0) / (n - 1) * i);
                        curveSamples.push(curve.evaluate([], t0));
                    }
                    curvePath.attr('d', lineGenerator(curveSamples));
                    break;
                case "BezierCurve2D" /* BEZIER */:
                    ctrlPath.attr('d', lineGenerator(this.controlPoints));
                    curvePath.attr('d', this.controlPoints.slice(1).reduce((d, p, i, a) => {
                        d += ` ${xScale(p[0])} ${yScale(p[1])}${i === a.length - 1 ? '' : ','}`;
                        return d;
                    }, `M ${xScale(this.controlPoints[0][0])} ${yScale(this.controlPoints[0][1])} C`));
                    break;
                case "Polyline2D" /* POLY_LINE */:
                    curvePath.attr('d', lineGenerator(this.controlPoints));
                    break;
                default:
                    break;
            }
            circles.data(this.controlPoints)
                .attr('cx', d => xScale(d[0]))
                .attr('cy', d => yScale(d[1]));
        };
        draw();
        const updateModel = debounce(points => {
            this.model.set('control_points', points);
            this.touch();
        }, 200);
        circles.call(// @ts-ignore
        d3.drag().on('drag', (d, i) => {
            const tx = d3.event.x + dx;
            const ty = d3.event.y + dy;
            if (tx > width - margin || tx < margin || ty > height - margin || ty < margin) {
                return;
            }
            const x = isFunction && (i === 0 || i === this.controlPoints.length - 1) ? this.controlPoints[i][0] : xScale.invert(d3.event.x);
            const y = yScale.invert(d3.event.y);
            this.controlPoints[i] = [x, y];
            draw();
            updateModel(this.controlPoints = this.controlPoints.slice());
        }));
    }
}
//# sourceMappingURL=editors.js.map